@@ -1,8 +1,8 @@
 /*
 **  OSSP sa - Socket Abstraction
-**  Copyright (c) 2001-2003 Ralf S. Engelschall <rse@engelschall.com>
-**  Copyright (c) 2001-2003 The OSSP Project <http://www.ossp.org/>
-**  Copyright (c) 2001-2003 Cable & Wireless Deutschland <http://www.cw.com/de/>
+**  Copyright (c) 2001-2004 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 2001-2004 The OSSP Project <http://www.ossp.org/>
+**  Copyright (c) 2001-2004 Cable & Wireless <http://www.cw.com/>
 **
 **  This file is part of OSSP sa, a socket abstraction library which
 **  can be found at http://www.ossp.org/pkg/lib/sa/.
@@ -116,49 +116,49 @@ const char sa_id[] = "OSSP sa";
 /* system call structure declaration macros */
 #define SA_SC_DECLARE_0(rc_t, fn) \
     struct { \
-        union { void (*any)(); \
+        union { void (*any)(void); \
                 rc_t (*std)(void); \
                 rc_t (*ctx)(void *); } fptr; \
         void *fctx; \
     } sc_##fn;
 #define SA_SC_DECLARE_1(rc_t, fn, a1_t) \
     struct { \
-        union { void (*any)(); \
+        union { void (*any)(void); \
                 rc_t (*std)(a1_t); \
                 rc_t (*ctx)(void *, a1_t); } fptr; \
         void *fctx; \
     } sc_##fn;
 #define SA_SC_DECLARE_2(rc_t, fn, a1_t, a2_t) \
     struct { \
-        union { void (*any)(); \
+        union { void (*any)(void); \
                 rc_t (*std)(a1_t, a2_t); \
                 rc_t (*ctx)(void *, a1_t, a2_t); } fptr; \
         void *fctx; \
     } sc_##fn;
 #define SA_SC_DECLARE_3(rc_t, fn, a1_t, a2_t, a3_t) \
     struct { \
-        union { void (*any)(); \
+        union { void (*any)(void); \
                 rc_t (*std)(a1_t, a2_t, a3_t); \
                 rc_t (*ctx)(void *, a1_t, a2_t, a3_t); } fptr; \
         void *fctx; \
     } sc_##fn;
 #define SA_SC_DECLARE_4(rc_t, fn, a1_t, a2_t, a3_t, a4_t) \
     struct { \
-        union { void (*any)(); \
+        union { void (*any)(void); \
                 rc_t (*std)(a1_t, a2_t, a3_t, a4_t); \
                 rc_t (*ctx)(void *, a1_t, a2_t, a3_t, a4_t); } fptr; \
         void *fctx; \
     } sc_##fn;
 #define SA_SC_DECLARE_5(rc_t, fn, a1_t, a2_t, a3_t, a4_t, a5_t) \
     struct { \
-        union { void (*any)(); \
+        union { void (*any)(void); \
                 rc_t (*std)(a1_t, a2_t, a3_t, a4_t, a5_t); \
                 rc_t (*ctx)(void *, a1_t, a2_t, a3_t, a4_t, a5_t); } fptr; \
         void *fctx; \
     } sc_##fn;
 #define SA_SC_DECLARE_6(rc_t, fn, a1_t, a2_t, a3_t, a4_t, a5_t, a6_t) \
     struct { \
-        union { void (*any)(); \
+        union { void (*any)(void); \
                 rc_t (*std)(a1_t, a2_t, a3_t, a4_t, a5_t, a6_t); \
                 rc_t (*ctx)(void *, a1_t, a2_t, a3_t, a4_t, a5_t, a6_t); } fptr; \
         void *fctx; \
@@ -167,7 +167,7 @@ const char sa_id[] = "OSSP sa";
 /* system call structure assignment macro */
 #define SA_SC_ASSIGN(sa, fn, ptr, ctx) \
     do { \
-        (sa)->scSysCall.sc_##fn.fptr.any = (void (*)())(ptr); \
+        (sa)->scSysCall.sc_##fn.fptr.any = (void (*)(void))(ptr); \
         (sa)->scSysCall.sc_##fn.fctx = (ctx); \
     } while (0)
 
@@ -182,7 +182,7 @@ const char sa_id[] = "OSSP sa";
 #define SA_SC_CALL_0(sa, fn) \
     (   (sa)->scSysCall.sc_##fn.fctx != NULL \
      ? ((sa)->scSysCall.sc_##fn.fptr.ctx)((sa)->scSysCall.sc_##fn.fctx) \
-     : ((sa)->scSysCall.sc_##fn.fptr.std)() )
+     : ((sa)->scSysCall.sc_##fn.fptr.std)(void) )
 #define SA_SC_CALL_1(sa, fn, a1) \
     (   (sa)->scSysCall.sc_##fn.fctx != NULL \
      ? ((sa)->scSysCall.sc_##fn.fptr.ctx)((sa)->scSysCall.sc_##fn.fctx, a1) \
@@ -939,14 +939,14 @@ static sa_rc_t sa_socket_settimeouts(sa_t *sa)
     if (sa->fdSocket == -1)
         return SA_OK;
 
-#if defined(SO_RCVTIMEO) && defined(SO_SNDTIMEO)
-    if (SA_TVISZERO(sa->tvTimeout[SA_TIMEOUT_READ])) {
+#if defined(SO_RCVTIMEO) && defined(USE_SO_RCVTIMEO) && defined(SO_SNDTIMEO) && defined(USE_SO_SNDTIMEO)
+    if (!SA_TVISZERO(sa->tvTimeout[SA_TIMEOUT_READ])) {
         if (setsockopt(sa->fdSocket, SOL_SOCKET, SO_RCVTIMEO,
                        (const void *)(&sa->tvTimeout[SA_TIMEOUT_READ]),
                        (socklen_t)(sizeof(sa->tvTimeout[SA_TIMEOUT_READ]))) < 0)
             return SA_RC(SA_ERR_SYS);
     }
-    if (SA_TVISZERO(sa->tvTimeout[SA_TIMEOUT_WRITE])) {
+    if (!SA_TVISZERO(sa->tvTimeout[SA_TIMEOUT_WRITE])) {
         if (setsockopt(sa->fdSocket, SOL_SOCKET, SO_SNDTIMEO,
                        (const void *)(&sa->tvTimeout[SA_TIMEOUT_WRITE]),
                        (socklen_t)(sizeof(sa->tvTimeout[SA_TIMEOUT_WRITE]))) < 0)
@@ -1393,7 +1393,7 @@ sa_rc_t sa_option(sa_t *sa, sa_option_t id, ...)
 }
 
 /* override system call */
-sa_rc_t sa_syscall(sa_t *sa, sa_syscall_t id, void (*fptr)(), void *fctx)
+sa_rc_t sa_syscall(sa_t *sa, sa_syscall_t id, void (*fptr)(void), void *fctx)
 {
     sa_rc_t rv;
 
@@ -1452,6 +1452,7 @@ sa_rc_t sa_connect(sa_t *sa, sa_addr_t *raddr)
     fd_set rset, wset;
     socklen_t len;
     sa_rc_t rv;
+    struct timeval *tv;
 
     /* argument sanity check(s) */
     if (sa == NULL || raddr == NULL)
@@ -1466,72 +1467,88 @@ sa_rc_t sa_connect(sa_t *sa, sa_addr_t *raddr)
         if ((rv = sa_socket_init(sa, raddr->nFamily)) != SA_OK)
             return SA_RC(rv);
 
+    /* prepare return code decision */
     rv = SA_OK;
-    if (SA_TVISZERO(sa->tvTimeout[SA_TIMEOUT_CONNECT])) {
-        /* standard/non-timeout-aware connect operation */
-        if (SA_SC_CALL_3(sa, connect, sa->fdSocket, raddr->saBuf, raddr->slBuf) < 0)
-            rv = SA_ERR_SYS;
-    }
-    else {
-        /* emulated/timeout-aware connect operation */
-        error = 0;
+    error = 0;
 
-        /* temporarily switch underlying socket to non-blocking mode */
+    /* temporarily switch underlying socket to non-blocking mode
+       (necessary under timeout-aware operation only) */
+    flags = 0;
+    if (!SA_TVISZERO(sa->tvTimeout[SA_TIMEOUT_CONNECT])) {
         flags = fcntl(sa->fdSocket, F_GETFL, 0);
         (void)fcntl(sa->fdSocket, F_SETFL, flags|O_NONBLOCK);
+    }
 
-        /* perform the connect operation */
-        if ((n = SA_SC_CALL_3(sa, connect, sa->fdSocket, raddr->saBuf, raddr->slBuf)) < 0) {
-            if (errno != EINPROGRESS) {
-                error = errno;
-                goto done;
-            }
-        }
-
-        /* ok if connect completed immediately */
-        if (n == 0)
-            goto done;
-
-        /* wait for read or write possibility */
-        FD_ZERO(&rset);
-        FD_ZERO(&wset);
-        FD_SET(sa->fdSocket, &rset);
-        FD_SET(sa->fdSocket, &wset);
-        do {
-            n = SA_SC_CALL_5(sa, select, sa->fdSocket+1, &rset, &wset, NULL,
-                             &sa->tvTimeout[SA_TIMEOUT_CONNECT]);
-        } while (n == -1 && errno == EINTR);
-
-        /* decide on return semantic */
-        if (n < 0) {
+    /* perform the connect operation */
+    if ((n = SA_SC_CALL_3(sa, connect, sa->fdSocket, raddr->saBuf, raddr->slBuf)) < 0) {
+        if (errno != EINTR && errno != EINPROGRESS) {
+            /* we have to perform the following post-processing under
+               EINPROGRESS anway, but actually also for EINTR according
+               to Unix Network Programming, volume 1, section 5.9, W.
+               Richard Stevens: "What we are doing [] is restarting
+               the interrupted system call ourself. This is fine for
+               accept, along with the functions such as read, write,
+               select and open. But there is one function that we cannot
+               restart ourself: connect. If this function returns EINTR,
+               we cannot call it again, as doing so will return an
+               immediate error. When connect is interrupted by a caught
+               signal and is not automatically restarted, we must call
+               select to wait for the connection to complete, as we
+               describe in section 15.3." */
             error = errno;
             goto done;
         }
-        else if (n == 0) {
-            (void)close(sa->fdSocket); /* stop TCP three-way handshake */
-            sa->fdSocket = -1;
-            rv = SA_ERR_TMT;
-            goto done;
-        }
-
-        /* fetch pending error */
-        len = (socklen_t)sizeof(error);
-        if (getsockopt(sa->fdSocket, SOL_SOCKET, SO_ERROR, (void *)&error, &len) < 0)
-            error = errno;
-
-        done:
-
-        /* reset socket flags */
-        (void)fcntl(sa->fdSocket, F_SETFL, flags);
-
-        /* optionally set errno */
-        if (error != 0) {
-            (void)close(sa->fdSocket); /* just in case */
-            sa->fdSocket = -1;
-            errno = error;
-            rv = SA_ERR_SYS;
-        }
     }
+
+    /* ok if connect completed immediately */
+    if (n == 0)
+        goto done;
+
+    /* wait for read or write possibility */
+    FD_ZERO(&rset);
+    FD_ZERO(&wset);
+    FD_SET(sa->fdSocket, &rset);
+    FD_SET(sa->fdSocket, &wset);
+    if (!SA_TVISZERO(sa->tvTimeout[SA_TIMEOUT_CONNECT]))
+        tv = &sa->tvTimeout[SA_TIMEOUT_CONNECT];
+    else
+        tv = NULL;
+    do {
+        n = SA_SC_CALL_5(sa, select, sa->fdSocket+1, &rset, &wset, NULL, tv);
+    } while (n == -1 && errno == EINTR);
+
+    /* decide on return semantic */
+    if (n < 0) {
+        error = errno;
+        goto done;
+    }
+    else if (n == 0) {
+        (void)close(sa->fdSocket); /* stop TCP three-way handshake */
+        sa->fdSocket = -1;
+        rv = SA_ERR_TMT;
+        goto done;
+    }
+
+    /* fetch pending error */
+    len = (socklen_t)sizeof(error);
+    if (getsockopt(sa->fdSocket, SOL_SOCKET, SO_ERROR, (void *)&error, &len) < 0)
+        error = errno;
+
+    done:
+
+    /* reset socket flags
+       (necessary under timeout-aware operation only) */
+    if (!SA_TVISZERO(sa->tvTimeout[SA_TIMEOUT_CONNECT]))
+        (void)fcntl(sa->fdSocket, F_SETFL, flags);
+
+    /* optionally set errno */
+    if (error != 0) {
+        (void)close(sa->fdSocket); /* just in case */
+        sa->fdSocket = -1;
+        errno = error;
+        rv = SA_ERR_SYS;
+    }
+
     return SA_RC(rv);
 }
 
@@ -1602,7 +1619,10 @@ sa_rc_t sa_accept(sa_t *sa, sa_addr_t **caddr, sa_t **csa)
 
     /* perform accept operation on underlying socket */
     sa_size = (socklen_t)sizeof(sa_buf);
-    if ((s = SA_SC_CALL_3(sa, accept, sa->fdSocket, (struct sockaddr *)&sa_buf, &sa_size)) == -1)
+    do {
+        s = SA_SC_CALL_3(sa, accept, sa->fdSocket, (struct sockaddr *)&sa_buf, &sa_size);
+    } while (s == -1 && errno == EINTR);
+    if (s == -1)
         return SA_RC(SA_ERR_SYS);
 
     /* create result address object */
@@ -1737,13 +1757,13 @@ sa_rc_t sa_getfd(sa_t *sa, int *fd)
 static int sa_read_raw(sa_t *sa, char *cpBuf, int nBufLen)
 {
     int rv;
-#if !(defined(SO_RCVTIMEO) && defined(SO_SNDTIMEO))
+#if !(defined(SO_RCVTIMEO) && defined(USE_SO_RCVTIMEO) && defined(SO_SNDTIMEO) && defined(USE_SO_SNDTIMEO))
     fd_set fds;
 #endif
 
     /* if timeout is enabled, perform explicit/smart blocking instead
        of implicitly/hard blocking in the read(2) system call */
-#if !(defined(SO_RCVTIMEO) && defined(SO_SNDTIMEO))
+#if !(defined(SO_RCVTIMEO) && defined(USE_SO_RCVTIMEO) && defined(SO_SNDTIMEO) && defined(USE_SO_SNDTIMEO))
     if (!SA_TVISZERO(sa->tvTimeout[SA_TIMEOUT_READ])) {
         FD_ZERO(&fds);
         FD_SET(sa->fdSocket, &fds);
@@ -1763,7 +1783,7 @@ static int sa_read_raw(sa_t *sa, char *cpBuf, int nBufLen)
         rv = (int)SA_SC_CALL_3(sa, read, sa->fdSocket, cpBuf, (size_t)nBufLen);
     } while (rv == -1 && errno == EINTR);
 
-#if defined(SO_RCVTIMEO) && defined(SO_SNDTIMEO)
+#if defined(SO_RCVTIMEO) && defined(USE_SO_RCVTIMEO) && defined(SO_SNDTIMEO) && defined(USE_SO_SNDTIMEO)
     if (rv == -1 && errno == EWOULDBLOCK)
         errno = ETIMEDOUT;
 #endif
@@ -1867,7 +1887,7 @@ sa_rc_t sa_read(sa_t *sa, char *cpBuf, size_t nBufReq, size_t *nBufRes)
     return SA_RC(rv);
 }
 
-/* read data from socket until [CR]LF (convinience function) */
+/* read data from socket until [CR]LF (convenience function) */
 sa_rc_t sa_readln(sa_t *sa, char *cpBuf, size_t nBufReq, size_t *nBufRes)
 {
     char c;
@@ -1916,13 +1936,13 @@ sa_rc_t sa_readln(sa_t *sa, char *cpBuf, size_t nBufReq, size_t *nBufRes)
 static int sa_write_raw(sa_t *sa, const char *cpBuf, int nBufLen)
 {
     int rv;
-#if !(defined(SO_RCVTIMEO) && defined(SO_SNDTIMEO))
+#if !(defined(SO_RCVTIMEO) && defined(USE_SO_RCVTIMEO) && defined(SO_SNDTIMEO) && defined(USE_SO_SNDTIMEO))
     fd_set fds;
 #endif
 
     /* if timeout is enabled, perform explicit/smart blocking instead
        of implicitly/hard blocking in the write(2) system call */
-#if !(defined(SO_RCVTIMEO) && defined(SO_SNDTIMEO))
+#if !(defined(SO_RCVTIMEO) && defined(USE_SO_RCVTIMEO) && defined(SO_SNDTIMEO) && defined(USE_SO_SNDTIMEO))
     if (!SA_TVISZERO(sa->tvTimeout[SA_TIMEOUT_WRITE])) {
         FD_ZERO(&fds);
         FD_SET(sa->fdSocket, &fds);
@@ -1942,7 +1962,7 @@ static int sa_write_raw(sa_t *sa, const char *cpBuf, int nBufLen)
         rv = (int)SA_SC_CALL_3(sa, write, sa->fdSocket, cpBuf, (size_t)nBufLen);
     } while (rv == -1 && errno == EINTR);
 
-#if defined(SO_RCVTIMEO) && defined(SO_SNDTIMEO)
+#if defined(SO_RCVTIMEO) && defined(USE_SO_RCVTIMEO) && defined(SO_SNDTIMEO) && defined(USE_SO_SNDTIMEO)
     if (rv == -1 && errno == EWOULDBLOCK)
         errno = ETIMEDOUT;
 #endif
@@ -2032,7 +2052,7 @@ static int sa_writef_cb(void *_ctx, const char *buffer, size_t bufsize)
     return (int)n;
 }
 
-/* write formatted string to socket (convinience function) */
+/* write formatted string to socket (convenience function) */
 sa_rc_t sa_writef(sa_t *sa, const char *cpFmt, ...)
 {
     va_list ap;
@@ -2249,7 +2269,7 @@ sa_rc_t sa_send(sa_t *sa, sa_addr_t *raddr, const char *buf, size_t buflen, size
     return SA_OK;
 }
 
-/* send formatted string to socket (convinience function) */
+/* send formatted string to socket (convenience function) */
 sa_rc_t sa_sendf(sa_t *sa, sa_addr_t *raddr, const char *cpFmt, ...)
 {
     va_list ap;
